@@ -67,6 +67,11 @@
 
 #define VIDEO_KEYSYM    "XF86Display"
 
+/* Number of seconds that the confirmation dialog will last before it resets the
+ * RANDR configuration to its old state.
+ */
+#define CONFIRMATION_DIALOG_SECONDS 30
+
 /* name of the icon files (gsd-xrandr.svg, etc.) */
 #define GSD_XRANDR_ICON_NAME "gsd-xrandr"
 
@@ -97,26 +102,304 @@ struct GsdXrandrManagerPrivate
         /* fn-F7 status */
         int             current_fn_f7_config;             /* -1 if no configs */
         GnomeRRConfig **fn_f7_configs;  /* NULL terminated, NULL if there are no configs */
+
+        /* Last time at which we got a "screen got reconfigured" event; see on_randr_event() */
+        guint32 last_config_timestamp;
 };
 
 static void     gsd_xrandr_manager_class_init  (GsdXrandrManagerClass *klass);
 static void     gsd_xrandr_manager_init        (GsdXrandrManager      *xrandr_manager);
 static void     gsd_xrandr_manager_finalize    (GObject             *object);
 
+static void error_message (GsdXrandrManager *mgr, const char *primary_text, GError *error_to_display, const char *secondary_text);
+
+static void status_icon_popup_menu (GsdXrandrManager *manager, guint button, guint32 timestamp);
+static void run_display_capplet (GtkWidget *widget);
+
 G_DEFINE_TYPE (GsdXrandrManager, gsd_xrandr_manager, G_TYPE_OBJECT)
 
 static gpointer manager_object = NULL;
 
+static void
+show_timestamps_dialog (GsdXrandrManager *manager, const char *msg)
+{
+#if 1
+        return;
+#else
+        struct GsdXrandrManagerPrivate *priv = manager->priv;
+        GtkWidget *dialog;
+        guint32 change_timestamp, config_timestamp;
+        static int serial;
 
-/* DBus method; see gsd-xrandr-manager.xml for the interface definition */
+        gnome_rr_screen_get_timestamps (priv->rw_screen, &change_timestamp, &config_timestamp);
+
+        dialog = gtk_message_dialog_new (NULL,
+                                         0,
+                                         GTK_MESSAGE_INFO,
+                                         GTK_BUTTONS_CLOSE,
+                                         "RANDR timestamps (%d):\n%s\nchange: %u\nconfig: %u",
+                                         serial++,
+                                         msg,
+                                         change_timestamp,
+                                         config_timestamp);
+        g_signal_connect (dialog, "response",
+                          G_CALLBACK (gtk_widget_destroy), NULL);
+        gtk_widget_show (dialog);
+#endif
+}
+
+/* Optionally filters out GNOME_RR_ERROR_NO_MATCHING_CONFIG from
+ * gnome_rr_config_apply_from_filename_with_time(), since that is not usually an error.
+ */
+static gboolean
+apply_configuration_from_filename (GsdXrandrManager *manager,
+                                   const char       *filename,
+                                   gboolean          no_matching_config_is_an_error,
+                                   guint32           timestamp,
+                                   GError          **error)
+{
+        struct GsdXrandrManagerPrivate *priv = manager->priv;
+        GError *my_error;
+        gboolean success;
+        char *str;
+
+        str = g_strdup_printf ("Applying %s with timestamp %d", filename, timestamp);
+        show_timestamps_dialog (manager, str);
+        g_free (str);
+
+        my_error = NULL;
+        success = gnome_rr_config_apply_from_filename_with_time (priv->rw_screen, filename, timestamp, &my_error);
+        if (success)
+                return TRUE;
+
+        if (g_error_matches (my_error, GNOME_RR_ERROR, GNOME_RR_ERROR_NO_MATCHING_CONFIG)) {
+                if (no_matching_config_is_an_error)
+                        goto fail;
+
+                /* This is not an error; the user probably changed his monitors
+                 * and so they don't match any of the stored configurations.
+                 */
+                g_error_free (my_error);
+                return TRUE;
+        }
+
+fail:
+        g_propagate_error (error, my_error);
+        return FALSE;
+}
+
+static void
+restore_backup_configuration_without_messages (const char *backup_filename, const char *intended_filename)
+{
+        backup_filename = gnome_rr_config_get_backup_filename ();
+        rename (backup_filename, intended_filename);
+}
+
+static void
+restore_backup_configuration (GsdXrandrManager *manager, const char *backup_filename, const char *intended_filename, guint32 timestamp)
+{
+        int saved_errno;
+
+        if (rename (backup_filename, intended_filename) == 0) {
+                GError *error;
+
+                error = NULL;
+                if (!apply_configuration_from_filename (manager, intended_filename, FALSE, timestamp, &error)) {
+                        error_message (manager, _("Could not restore the display's configuration"), error, NULL);
+
+                        if (error)
+                                g_error_free (error);
+                }
+
+                return;
+        }
+
+        saved_errno = errno;
+
+        /* ENOENT means the original file didn't exist.  That is *not* an error;
+         * the backup was not created because there wasn't even an original
+         * monitors.xml (such as on a first-time login).  Note that *here* there
+         * is a "didn't work" monitors.xml, so we must delete that one.
+         */
+        if (saved_errno == ENOENT)
+                unlink (intended_filename);
+        else {
+                char *msg;
+
+                msg = g_strdup_printf ("Could not rename %s to %s: %s",
+                                       backup_filename, intended_filename,
+                                       g_strerror (saved_errno));
+                error_message (manager,
+                               _("Could not restore the display's configuration from a backup"),
+                               NULL,
+                               msg);
+                g_free (msg);
+        }
+
+        unlink (backup_filename);
+}
+
+typedef struct {
+        GsdXrandrManager *manager;
+        GtkWidget *dialog;
+
+        int countdown;
+        int response_id;
+} TimeoutDialog;
+
+static void
+print_countdown_text (TimeoutDialog *timeout)
+{
+        gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (timeout->dialog),
+                                                  ngettext ("The display will be reset to its previous configuration in %d second",
+                                                            "The display will be reset to its previous configuration in %d seconds",
+                                                            timeout->countdown),
+                                                  timeout->countdown);
+}
+
+static gboolean
+timeout_cb (gpointer data)
+{
+        TimeoutDialog *timeout = data;
+
+        timeout->countdown--;
+
+        if (timeout->countdown == 0) {
+                timeout->response_id = GTK_RESPONSE_CANCEL;
+                gtk_main_quit ();
+        } else {
+                print_countdown_text (timeout);
+        }
+
+        return TRUE;
+}
+
+static void
+timeout_response_cb (GtkDialog *dialog, int response_id, gpointer data)
+{
+        TimeoutDialog *timeout = data;
+
+        if (response_id == GTK_RESPONSE_DELETE_EVENT) {
+                /* The user closed the dialog or pressed ESC, revert */
+                timeout->response_id = GTK_RESPONSE_CANCEL;
+        } else
+                timeout->response_id = response_id;
+
+        gtk_main_quit ();
+}
+
+static gboolean
+user_says_things_are_ok (GsdXrandrManager *manager, GdkWindow *parent_window)
+{
+        TimeoutDialog timeout;
+        guint timeout_id;
+
+        timeout.manager = manager;
+
+        /* It sucks that we don't know the parent window here */
+        timeout.dialog = gtk_message_dialog_new (NULL,
+                                                 GTK_DIALOG_MODAL,
+                                                 GTK_MESSAGE_QUESTION,
+                                                 GTK_BUTTONS_NONE,
+                                                 _("Does the display look OK?"));
+
+        timeout.countdown = CONFIRMATION_DIALOG_SECONDS;
+
+        print_countdown_text (&timeout);
+
+        gtk_dialog_add_button (GTK_DIALOG (timeout.dialog), _("_Restore Previous Configuration"), GTK_RESPONSE_CANCEL);
+        gtk_dialog_add_button (GTK_DIALOG (timeout.dialog), _("_Keep This Configuration"), GTK_RESPONSE_ACCEPT);
+        gtk_dialog_set_default_response (GTK_DIALOG (timeout.dialog), GTK_RESPONSE_ACCEPT); /* ah, the optimism */
+
+        g_signal_connect (timeout.dialog, "response",
+                          G_CALLBACK (timeout_response_cb),
+                          &timeout);
+
+        gtk_widget_realize (timeout.dialog);
+
+        if (parent_window)
+                gdk_window_set_transient_for (gtk_widget_get_window (timeout.dialog), parent_window);
+
+        gtk_widget_show_all (timeout.dialog);
+        /* We don't use g_timeout_add_seconds() since we actually care that the user sees "real" second ticks in the dialog */
+        timeout_id = g_timeout_add (1000,
+                                    timeout_cb,
+                                    &timeout);
+        gtk_main ();
+
+        gtk_widget_destroy (timeout.dialog);
+        g_source_remove (timeout_id);
+
+        if (timeout.response_id == GTK_RESPONSE_ACCEPT)
+                return TRUE;
+        else
+                return FALSE;
+}
+
+static gboolean
+try_to_apply_intended_configuration (GsdXrandrManager *manager, GdkWindow *parent_window, guint32 timestamp, GError **error)
+{
+        char *backup_filename;
+        char *intended_filename;
+        gboolean result;
+
+        /* Try to apply the intended configuration */
+
+        backup_filename = gnome_rr_config_get_backup_filename ();
+        intended_filename = gnome_rr_config_get_intended_filename ();
+
+        result = apply_configuration_from_filename (manager, intended_filename, FALSE, timestamp, error);
+        if (!result) {
+                error_message (manager, _("The selected configuration for displays could not be applied"), error ? *error : NULL, NULL);
+                restore_backup_configuration_without_messages (backup_filename, intended_filename);
+                goto out;
+        }
+
+        /* Confirm with the user */
+
+        if (user_says_things_are_ok (manager, parent_window))
+                unlink (backup_filename);
+        else
+                restore_backup_configuration (manager, backup_filename, intended_filename, timestamp);
+
+out:
+        g_free (backup_filename);
+        g_free (intended_filename);
+
+        return result;
+}
+
+/* DBus method for org.gnome.SettingsDaemon.XRANDR ApplyConfiguration; see gsd-xrandr-manager.xml for the interface definition */
 static gboolean
 gsd_xrandr_manager_apply_configuration (GsdXrandrManager *manager,
                                         GError          **error)
 {
-        GnomeRRScreen *screen = manager->priv->rw_screen;
-
-        return gnome_rr_config_apply_stored (screen, error);
+        return try_to_apply_intended_configuration (manager, NULL, GDK_CURRENT_TIME, error);
 }
+
+/* DBus method for org.gnome.SettingsDaemon.XRANDR_2 ApplyConfiguration; see gsd-xrandr-manager.xml for the interface definition */
+static gboolean
+gsd_xrandr_manager_2_apply_configuration (GsdXrandrManager *manager,
+                                          gint64            parent_window_id,
+                                          gint64            timestamp,
+                                          GError          **error)
+{
+        GdkWindow *parent_window;
+        gboolean result;
+
+        if (parent_window_id != 0)
+                parent_window = gdk_window_foreign_new_for_display (gdk_display_get_default (), (GdkNativeWindow) parent_window_id);
+        else
+                parent_window = NULL;
+
+        result = try_to_apply_intended_configuration (manager, parent_window, (guint32) timestamp, error);
+
+        if (parent_window)
+                g_object_unref (parent_window);
+
+        return result;
+}
+
 /* We include this after the definition of gsd_xrandr_manager_apply_configuration() so the prototype will already exist */
 #include "gsd-xrandr-manager-glue.h"
 
@@ -378,7 +661,7 @@ sanitize (GPtrArray *array)
         int i;
         GPtrArray *new;
 
-        g_print ("before sanitizing\n");
+        g_debug ("before sanitizing");
 
         for (i = 0; i < array->len; ++i) {
                 if (array->pdata[i]) {
@@ -390,7 +673,7 @@ sanitize (GPtrArray *array)
         /* Remove configurations that are duplicates of
          * configurations earlier in the cycle
          */
-        for (i = 0; i < array->len; ++i) {
+        for (i = 1; i < array->len; ++i) {
                 int j;
 
                 for (j = 0; j < i; ++j) {
@@ -398,7 +681,7 @@ sanitize (GPtrArray *array)
                         GnomeRRConfig *other = array->pdata[i];
 
                         if (this && other && gnome_rr_config_equal (this, other)) {
-                                g_print ("removing duplicate configuration\n");
+                                g_debug ("removing duplicate configuration");
                                 gnome_rr_config_free (this);
                                 array->pdata[j] = NULL;
                                 break;
@@ -435,7 +718,12 @@ sanitize (GPtrArray *array)
                 }
         }
 
-        g_ptr_array_add (new, NULL);
+        if (new->len > 0) {
+                g_ptr_array_add (new, NULL);
+        } else {
+                g_ptr_array_free (new, TRUE);
+                new = NULL;
+        }
 
         g_ptr_array_free (array, TRUE);
 
@@ -448,7 +736,7 @@ generate_fn_f7_configs (GsdXrandrManager *mgr)
         GPtrArray *array = g_ptr_array_new ();
         GnomeRRScreen *screen = mgr->priv->rw_screen;
 
-        g_print ("Generating configurations\n");
+        g_debug ("Generating configurations");
 
         /* Free any existing list of configurations */
         if (mgr->priv->fn_f7_configs) {
@@ -463,17 +751,18 @@ generate_fn_f7_configs (GsdXrandrManager *mgr)
         }
 
         g_ptr_array_add (array, gnome_rr_config_new_current (screen));
-        g_ptr_array_add (array, make_xinerama_setup (screen));
         g_ptr_array_add (array, make_clone_setup (screen));
+        g_ptr_array_add (array, make_xinerama_setup (screen));
         g_ptr_array_add (array, make_laptop_setup (screen));
         g_ptr_array_add (array, make_other_setup (screen));
         g_ptr_array_add (array, gnome_rr_config_new_stored (screen, NULL)); /* NULL-GError - if this can't read the stored config, no big deal */
-        g_ptr_array_add (array, NULL);
 
         array = sanitize (array);
 
-        mgr->priv->fn_f7_configs = (GnomeRRConfig **)g_ptr_array_free (array, FALSE);
-        mgr->priv->current_fn_f7_config = 0;
+        if (array) {
+                mgr->priv->fn_f7_configs = (GnomeRRConfig **)g_ptr_array_free (array, FALSE);
+                mgr->priv->current_fn_f7_config = 0;
+        }
 }
 
 static void
@@ -485,10 +774,17 @@ error_message (GsdXrandrManager *mgr, const char *primary_text, GError *error_to
 
         g_assert (error_to_display == NULL || secondary_text == NULL);
 
-        notification = notify_notification_new_with_status_icon (primary_text,
-                                                                 error_to_display ? error_to_display->message : secondary_text,
-                                                                 GSD_XRANDR_ICON_NAME,
-                                                                 priv->status_icon);
+        if (priv->status_icon)
+                notification = notify_notification_new_with_status_icon (primary_text,
+                                                                         error_to_display ? error_to_display->message : secondary_text,
+                                                                         GSD_XRANDR_ICON_NAME,
+                                                                         priv->status_icon);
+        else
+                notification = notify_notification_new (primary_text,
+                                                        error_to_display ? error_to_display->message : secondary_text,
+                                                        GSD_XRANDR_ICON_NAME,
+                                                        NULL);
+
         notify_notification_show (notification, NULL); /* NULL-GError */
 #else
         GtkWidget *dialog;
@@ -504,7 +800,7 @@ error_message (GsdXrandrManager *mgr, const char *primary_text, GError *error_to
 }
 
 static void
-handle_fn_f7 (GsdXrandrManager *mgr)
+handle_fn_f7 (GsdXrandrManager *mgr, guint32 timestamp)
 {
         GsdXrandrManagerPrivate *priv = mgr->priv;
         GnomeRRScreen *screen = priv->rw_screen;
@@ -526,7 +822,7 @@ handle_fn_f7 (GsdXrandrManager *mgr)
         g_debug ("Handling fn-f7");
 
         error = NULL;
-        if (!gnome_rr_screen_refresh (screen, &error)) {
+        if (!gnome_rr_screen_refresh (screen, &error) && error) {
                 char *str;
 
                 str = g_strdup_printf (_("Could not refresh the screen information: %s"), error->message);
@@ -565,7 +861,7 @@ handle_fn_f7 (GsdXrandrManager *mgr)
                 g_debug ("applying");
 
                 error = NULL;
-                if (!gnome_rr_config_apply (priv->fn_f7_configs[mgr->priv->current_fn_f7_config], screen, &error)) {
+                if (!gnome_rr_config_apply_with_time (priv->fn_f7_configs[mgr->priv->current_fn_f7_config], screen, timestamp, &error)) {
                         error_message (mgr, _("Could not switch the monitor configuration"), error, NULL);
                         g_error_free (error);
                 }
@@ -591,8 +887,8 @@ event_filter (GdkXEvent           *xevent,
         if (xev->xany.type != KeyPress && xev->xany.type != KeyRelease)
                 return GDK_FILTER_CONTINUE;
 
-        if (xev->xkey.keycode == manager->priv->keycode && xev->xany.type == KeyPress) {
-                handle_fn_f7 (manager);
+        if (xev->xany.type == KeyPress && xev->xkey.keycode == manager->priv->keycode) {
+                handle_fn_f7 (manager, xev->xkey.time);
 
                 return GDK_FILTER_CONTINUE;
         }
@@ -601,23 +897,237 @@ event_filter (GdkXEvent           *xevent,
 }
 
 static void
-on_randr_event (GnomeRRScreen *screen, gpointer data)
+refresh_tray_icon_menu_if_active (GsdXrandrManager *manager, guint32 timestamp)
 {
-        GsdXrandrManager *manager = GSD_XRANDR_MANAGER (data);
+        GsdXrandrManagerPrivate *priv = manager->priv;
 
-        if (!manager->priv->running)
-                return;
-
-        /* FIXME: Set up any new screens here */
+        if (priv->popup_menu) {
+                gtk_menu_shell_cancel (GTK_MENU_SHELL (priv->popup_menu)); /* status_icon_popup_menu_selection_done_cb() will free everything */
+                status_icon_popup_menu (manager, 0, timestamp);
+        }
 }
 
 static void
-popup_menu_configure_display_cb (GtkMenuItem *item, gpointer data)
+auto_configure_outputs (GsdXrandrManager *manager, guint32 timestamp)
+{
+        GsdXrandrManagerPrivate *priv = manager->priv;
+        GnomeRRConfig *config;
+        int i;
+        GList *just_turned_on;
+        GList *l;
+        int x;
+        GError *error;
+
+        config = gnome_rr_config_new_current (priv->rw_screen);
+
+        /* For outputs that are connected and on (i.e. they have a CRTC assigned
+         * to them, so they are getting a signal), we leave them as they are
+         * with their current modes.
+         *
+         * For other outputs, we will turn on connected-but-off outputs and turn
+         * off disconnected-but-on outputs.
+         *
+         * FIXME: If an output remained connected+on, it would be nice to ensure
+         * that the output's CRTCs still has a reasonable mode (think of
+         * changing one monitor for another with different capabilities).
+         */
+
+        just_turned_on = NULL;
+
+        for (i = 0; config->outputs[i] != NULL; i++) {
+                GnomeOutputInfo *output = config->outputs[i];
+
+                if (output->connected && !output->on) {
+                        output->on = TRUE;
+                        output->rotation = GNOME_RR_ROTATION_0;
+                        just_turned_on = g_list_prepend (just_turned_on, GINT_TO_POINTER (i));
+                } else if (!output->connected && output->on)
+                        output->on = FALSE;
+        }
+
+        /* Now, lay out the outputs from left to right.  Put first the outputs
+         * which remained on; put last the outputs that were newly turned on.
+         */
+
+        x = 0;
+
+        /* First, outputs that remained on */
+
+        for (i = 0; config->outputs[i] != NULL; i++) {
+                GnomeOutputInfo *output = config->outputs[i];
+
+                if (g_list_find (just_turned_on, GINT_TO_POINTER (i)))
+                        continue;
+
+                if (output->on) {
+                        g_assert (output->connected);
+
+                        output->x = x;
+                        output->y = 0;
+
+                        x += output->width;
+                }
+        }
+
+        /* Second, outputs that were newly-turned on */
+
+        for (l = just_turned_on; l; l = l->next) {
+                GnomeOutputInfo *output;
+
+                i = GPOINTER_TO_INT (l->data);
+                output = config->outputs[i];
+
+                g_assert (output->on && output->connected);
+
+                output->x = x;
+                output->y = 0;
+
+                /* since the output was off, use its preferred width/height (it doesn't have a real width/height yet) */
+                output->width = output->pref_width;
+                output->height = output->pref_height;
+
+                x += output->width;
+        }
+
+        /* Apply the configuration! */
+
+        error = NULL;
+        if (!gnome_rr_config_apply_with_time (config, priv->rw_screen, timestamp, &error)) {
+                error_message (manager, _("Could not switch the monitor configuration"), error, NULL);
+                g_error_free (error);
+        }
+
+        g_list_free (just_turned_on);
+        gnome_rr_config_free (config);
+
+        /* Finally, even though we did a best-effort job in sanitizing the
+         * outputs, we don't know the physical layout of the monitors.  We'll
+         * start the display capplet so that the user can tweak things to his
+         * liking.
+         */
+
+#if 0
+        /* FIXME: This is disabled for now.  The capplet is not a single-instance application.
+         * If you do this:
+         *
+         *   1. Start the display capplet
+         *
+         *   2. Plug an extra monitor
+         *
+         *   3. Hit the "Detect displays" button
+         *
+         * Then we will get a RANDR event because X re-probes the outputs.  We don't want to
+         * start up a second display capplet right there!
+         */
+
+        run_display_capplet (NULL);
+#endif
+}
+
+static void
+on_randr_event (GnomeRRScreen *screen, gpointer data)
+{
+        GsdXrandrManager *manager = GSD_XRANDR_MANAGER (data);
+        GsdXrandrManagerPrivate *priv = manager->priv;
+        guint32 change_timestamp, config_timestamp;
+
+        if (!priv->running)
+                return;
+
+        gnome_rr_screen_get_timestamps (screen, &change_timestamp, &config_timestamp);
+
+        if (change_timestamp >= config_timestamp) {
+                /* The event is due to an explicit configuration change.
+                 *
+                 * If the change was performed by us, then we need to do nothing.
+                 *
+                 * If the change was done by some other X client, we don't need
+                 * to do anything, either; the screen is already configured.
+                 */
+                show_timestamps_dialog (manager, "ignoring since change > config");
+        } else {
+                /* Here, config_timestamp > change_timestamp.  This means that
+                 * the screen got reconfigured because of hotplug/unplug; the X
+                 * server is just notifying us, and we need to configure the
+                 * outputs in a sane way.
+                 */
+
+
+#if 1
+                if (config_timestamp != priv->last_config_timestamp) {
+                        priv->last_config_timestamp = config_timestamp;
+                        auto_configure_outputs (manager, config_timestamp);
+                }
+#else
+                /* WHY THIS CODE IS DISABLED:
+                 *
+                 * The strategy of "on hotplug or unsuspend, restore a
+                 * known-good configuration, and fall back to autoconfiguration"
+                 * works fine as long as you don't happen to be running
+                 * gnome-display-properties and click its "Detect displays"
+                 * button.
+                 *
+                 * If you do that, the RANDR calls from g-d-p will cause the X
+                 * server to re-probe the RANDR outputs.  The server will send
+                 * us an event, we'll restore the configuration to something
+                 * else... and you'll be weirded out, because "just detecting
+                 * your monitors" should not change the current configuration,
+                 * right?
+                 *
+                 * We may need some kind of D-bus API so that g-d-p can inhibit
+                 * this RANDR plugin's reconfiguration-fu when the "Detect
+                 * displays" button is being used.
+                 */
+
+                char *intended_filename;
+                GError *error;
+                gboolean success;
+
+                show_timestamps_dialog (manager, "need to deal with reconfiguration, as config >= change");
+
+                intended_filename = gnome_rr_config_get_intended_filename ();
+
+                error = NULL;
+                success = apply_configuration_from_filename (manager, intended_filename, TRUE, config_timestamp, &error);
+                g_free (intended_filename);
+
+                if (!success) {
+                        /* We don't bother checking the error type.
+                         *
+                         * Both G_FILE_ERROR_NOENT and
+                         * GNOME_RR_ERROR_NO_MATCHING_CONFIG would mean, "there
+                         * was no configuration to apply, or none that matched
+                         * the current outputs", and in that case we need to run
+                         * our fallback.
+                         *
+                         * Any other error means "we couldn't do the smart thing
+                         * of using a previously- saved configuration, anyway,
+                         * for some other reason.  In that case, we also need to
+                         * run our fallback to avoid leaving the user with a
+                         * bogus configuration.
+                         */
+
+                        if (error)
+                                g_error_free (error);
+
+                        auto_configure_outputs (manager, config_timestamp);
+                }
+#endif
+        }
+
+        refresh_tray_icon_menu_if_active (manager, MAX (change_timestamp, config_timestamp));
+}
+
+static void
+run_display_capplet (GtkWidget *widget)
 {
         GdkScreen *screen;
         GError *error;
 
-        screen = gtk_widget_get_screen (GTK_WIDGET (item));
+        if (widget)
+                screen = gtk_widget_get_screen (widget);
+        else
+                screen = gdk_screen_get_default ();
 
         error = NULL;
         if (!gdk_spawn_command_line_on_screen (screen, GSD_XRANDR_DISPLAY_CAPPLET, &error)) {
@@ -635,6 +1145,12 @@ popup_menu_configure_display_cb (GtkMenuItem *item, gpointer data)
         }
 }
 
+static void
+popup_menu_configure_display_cb (GtkMenuItem *item, gpointer data)
+{
+        run_display_capplet (GTK_WIDGET (item));
+}
+
 static void
 status_icon_popup_menu_selection_done_cb (GtkMenuShell *menu_shell, gpointer data)
 {
@@ -772,6 +1288,7 @@ make_menu_item_for_output_title (GsdXrandrManager *manager, GnomeOutputInfo *out
         GtkWidget *item;
         GtkWidget *label;
         char *str;
+	GdkColor black = { 0, 0, 0, 0 };
 
         item = gtk_menu_item_new ();
 
@@ -783,6 +1300,12 @@ make_menu_item_for_output_title (GsdXrandrManager *manager, GnomeOutputInfo *out
         gtk_label_set_markup (GTK_LABEL (label), str);
         g_free (str);
 
+	/* Make the label explicitly black.  We don't want it to follow the
+	 * theme's colors, since the label is always shown against a light
+	 * pastel background.  See bgo#556050
+	 */
+	gtk_widget_modify_fg (label, GTK_WIDGET_STATE (label), &black);
+
         /* Add padding around the label to fit the box that we'll draw for color-coding */
         gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
         gtk_misc_set_padding (GTK_MISC (label),
@@ -869,6 +1392,31 @@ add_unsupported_rotation_item (GsdXrandrManager *manager)
         gtk_menu_shell_append (GTK_MENU_SHELL (priv->popup_menu), item);
 }
 
+static void
+ensure_current_configuration_is_saved (void)
+{
+        GnomeRRScreen *rr_screen;
+        GnomeRRConfig *rr_config;
+
+        /* Normally, gnome_rr_config_save() creates a backup file based on the
+         * old monitors.xml.  However, if *that* file didn't exist, there is
+         * nothing from which to create a backup.  So, here we'll save the
+         * current/unchanged configuration and then let our caller call
+         * gnome_rr_config_save() again with the new/changed configuration, so
+         * that there *will* be a backup file in the end.
+         */
+
+        rr_screen = gnome_rr_screen_new (gdk_screen_get_default (), NULL, NULL, NULL); /* NULL-GError */
+        if (!rr_screen)
+                return;
+
+        rr_config = gnome_rr_config_new_current (rr_screen);
+        gnome_rr_config_save (rr_config, NULL); /* NULL-GError */
+
+        gnome_rr_config_free (rr_config);
+        gnome_rr_screen_destroy (rr_screen);
+}
+
 static void
 output_rotation_item_activate_cb (GtkCheckMenuItem *item, gpointer data)
 {
@@ -882,6 +1430,8 @@ output_rotation_item_activate_cb (GtkCheckMenuItem *item, gpointer data)
 	if (!gtk_check_menu_item_get_active (item))
 		return;
 
+        ensure_current_configuration_is_saved ();
+
         output = g_object_get_data (G_OBJECT (item), "output");
         rotation = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (item), "rotation"));
 
@@ -896,13 +1446,7 @@ output_rotation_item_activate_cb (GtkCheckMenuItem *item, gpointer data)
                 return;
         }
 
-        if (!gnome_rr_config_apply_stored (priv->rw_screen, &error)) {
-                error_message (manager, _("The selected rotation could not be applied"), error, NULL);
-                if (error)
-                        g_error_free (error);
-
-                return;
-        }
+        try_to_apply_intended_configuration (manager, NULL, gtk_get_current_event_time (), NULL); /* NULL-GError */
 }
 
 static void
@@ -1076,7 +1620,7 @@ status_icon_start (GsdXrandrManager *manager)
          */
         if (!priv->status_icon) {
                 priv->status_icon = gtk_status_icon_new_from_icon_name (GSD_XRANDR_ICON_NAME);
-                gtk_status_icon_set_tooltip (priv->status_icon, _("Configure display settings"));
+                gtk_status_icon_set_tooltip_text (priv->status_icon, _("Configure display settings"));
 
                 g_signal_connect (priv->status_icon, "activate",
                                   G_CALLBACK (status_icon_activate_cb), manager);
@@ -1121,12 +1665,81 @@ on_config_changed (GConfClient          *client,
         start_or_stop_icon (manager);
 }
 
+static void
+apply_intended_configuration (GsdXrandrManager *manager, const char *intended_filename, guint32 timestamp)
+{
+        GError *my_error;
+
+        my_error = NULL;
+        if (!apply_configuration_from_filename (manager, intended_filename, FALSE, timestamp, &my_error)) {
+                if (my_error) {
+                        if (!g_error_matches (my_error, G_FILE_ERROR, G_FILE_ERROR_NOENT))
+                                error_message (manager, _("Could not apply the stored configuration for monitors"), my_error, NULL);
+
+                        g_error_free (my_error);
+                }
+        }
+}
+
+static void
+apply_stored_configuration_at_startup (GsdXrandrManager *manager, guint32 timestamp)
+{
+        GError *my_error;
+        gboolean success;
+        char *backup_filename;
+        char *intended_filename;
+
+        backup_filename = gnome_rr_config_get_backup_filename ();
+        intended_filename = gnome_rr_config_get_intended_filename ();
+
+        /* 1. See if there was a "saved" configuration.  If there is one, it means
+         * that the user had selected to change the display configuration, but the
+         * machine crashed.  In that case, we'll apply *that* configuration and save it on top of the
+         * "intended" one.
+         */
+
+        my_error = NULL;
+
+        success = apply_configuration_from_filename (manager, backup_filename, FALSE, timestamp, &my_error);
+        if (success) {
+                /* The backup configuration existed, and could be applied
+                 * successfully, so we must restore it on top of the
+                 * failed/intended one.
+                 */
+                restore_backup_configuration (manager, backup_filename, intended_filename, timestamp);
+                goto out;
+        }
+
+        if (!g_error_matches (my_error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
+                /* Epic fail:  there (probably) was a backup configuration, but
+                 * we could not apply it.  The only thing we can do is delete
+                 * the backup configuration.  Let's hope that the user doesn't
+                 * get left with an unusable display...
+                 */
+
+                unlink (backup_filename);
+                goto out;
+        }
+
+        /* 2. There was no backup configuration!  This means we are
+         * good.  Apply the intended configuration instead.
+         */
+
+        apply_intended_configuration (manager, intended_filename, timestamp);
+
+out:
+
+        if (my_error)
+                g_error_free (my_error);
+
+        g_free (backup_filename);
+        g_free (intended_filename);
+}
+
 gboolean
 gsd_xrandr_manager_start (GsdXrandrManager *manager,
                           GError          **error)
 {
-        GError *my_error;
-
         g_debug ("Starting xrandr manager");
         gnome_settings_profile_start (NULL);
 
@@ -1163,14 +1776,8 @@ gsd_xrandr_manager_start (GsdXrandrManager *manager,
                 gdk_error_trap_pop ();
         }
 
-        my_error = NULL;
-        if (!gnome_rr_config_apply_stored (manager->priv->rw_screen, &my_error)) {
-                /* my_error can be null if there were no stored configurations */
-                if (my_error) {
-                        error_message (manager, _("Could not apply the stored configuration for monitors"), my_error, NULL);
-                        g_error_free (my_error);
-                }
-        }
+        show_timestamps_dialog (manager, "Startup");
+        apply_stored_configuration_at_startup (manager, GDK_CURRENT_TIME); /* we don't have a real timestamp at startup anyway */
 
         gdk_window_add_filter (gdk_get_default_root_window(),
                                (GdkFilterFunc)event_filter,
